/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/
/**
 * @file applications/afs/encoding/downloadutil.c
 * @author Christian Grothoff
 *
 * Download helper methods (which do the real work).
 **/

#include "config.h"
#include "applications/afs/encoding/downloadutil.h"
#include "applications/afs/util/contentencoding.h"
#include "applications/afs/encoding/requestmanager.h"
#include "applications/afs/encoding/block.h"  
#include "util/cron.h"
#include "util/configuration.h"
#include "util/checksum.h"
#include "util/logging.h"
#include <unistd.h>

typedef struct PMWrap {
  ProgressModel userModel;
  void * userData;
  NodeContext * nc;
} PMWrap;

static void pModelWrap(ProgressStats * stats,
		       PMWrap * wrap) {
  if (wrap->userModel != NULL) {
    wrap->userModel(stats, wrap->userData);
  }
  if (stats->progress == stats->filesize) {    
    if (stats->progress == 0)
      freeIOC(&wrap->nc->ioc,
	      NO);
    else
      freeIOC(&wrap->nc->ioc, 
	      YES);
    xfree(wrap->nc, "freeing node context");
    xfree(wrap, "freeing PMWrap");
  }
}

/**
 * Download a file.
 * @param topCHK the CHK of the topmost IBlock
 * @param crc32 the CRC of the file
 * @param filesize the size of the file
 * @param fileName the name of the file
 * @param model the download model used to
 *        update status information; points to NULL if
 *        no status updates shall be given, otherwise 
 *        to a method that takes two size_t arguments
 *        (retrieved so far, total).
 * @param data pointer that is passed to the model method
 * @return a request manager that can be used to abort on 
 *         success, NULL on error
 **/
RequestManager * downloadFile(CHK_Hashes * topCHK,
			      int crc32,
			      size_t filesize,
			      char * fileName,
			      ProgressModel model,
			      void * data) {
  NodeContext * nc;
  Block * top;
  RequestManager * rm;
  struct PMWrap * wrap;

  nc = xmalloc(sizeof(NodeContext),
	       "allocating node context");
  if ( (rm = createRequestManager()) == NULL )
    return NULL;

  if (SYSERR == createIOContext(&nc->ioc,
				filesize,
				fileName,
				NO)) {
    destroyRequestManager(rm);
    return NULL;
  }
  wrap = xmalloc(sizeof(PMWrap), "wrapped pm");
  wrap->userModel = model;
  wrap->userData = data;
  wrap->nc = nc;
  nc->priority = 0; /* unused */
  nc->index = 0; /* unused */
  nc->pmodel = (ProgressModel) &pModelWrap;
  nc->data = wrap;
  memset(&nc->stats, 0, sizeof(ProgressStats));
  nc->stats.filesize = filesize;

  if (filesize <= sizeof(CONTENT_Block))
    top = (Block*) createTopDBlock(filesize);
  else
    top = (Block*) createTopIBlock(filesize);
  memcpy(&top->chk,
	 topCHK,
	 sizeof(CHK_Hashes));
  top->topCrc32 = crc32;
  top->download(top, nc, rm);
  return rm;
}
