/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/
/**
 * This file provides methods to manipulate host credit
 * rankings. It should be merged with outbuffer soon.
 * @author Tzvetan Horozov
 * @author Christian Grothoff
 **/ 

#ifndef CONNECTION_H
#define CONNECTION_H

#include "config.h"
#include "util/symcipher.h"
#include "util/configuration.h"
#include "server/policy.h"
#include "util/ports.h"
#include "util/cron.h"
#include "util/semaphore.h"
#include "util/checksum.h"

/* ********************* some message types ************** */

/**
 * Format of a timestamp-message. The rest of the body
 * is only valid if the timestamp is greater than the
 * current time (in seconds after 1970...).
 * Used against replay attacks!
 **/
typedef struct {
  p2p_HEADER header;
  /* timestamp  (network byte order)*/
  time_t timeStamp;
} TIMESTAMP_Message;

/**
 * Sequence number. If the sequence number is lower than
 * a previous number, the rest of the packet should be
 * ignored (replay). This will of course break if UDP packets
 * arrive out-of-order, but this is rare and we're best-effort.
 * This is used to defend against replay-attacks.
 **/
typedef struct {
  p2p_HEADER header;
  /* sequence number, in network byte order */
  unsigned int sequenceNumber;
} SEQUENCE_Message;

/**
 * The other side has decided to terminate the connection.
 * This message MAY be send if the other node decides to 
 * be nice. It is not required. Mind that the message
 * contains for which host the termination is, such
 * that we don't hang up the wrong connection...
 * A node can also choose to ignore the HANGUP message,
 * though this is probably not going to help that node.
 * This message is used to prevent sending data to 
 * connections that were closed on the other side (can
 * happen anyway, so this is just an optimization between
 * well-behaved, non-malicious nodes that like each other).
 **/
typedef struct {
  p2p_HEADER header;
  HostIdentity sender;
} HANGUP_Message;

/* ********************** Exported globals ******************************* */

extern char ** CONNECTION_knownHosts;
extern int CONNECTION_currentMaxKnownHosts;
extern int CONNECTION_currentKnownHosts;
extern Mutex CONNECTION_knownHostsLock;
extern char * CONNECTION_hostsDirectory;
extern HostIdentity myIdentity;

/* ***************** status constants ********** */

#define STAT_DOWN             0
#define STAT_WAITING_FOR_PING 1
#define STAT_WAITING_FOR_PONG 2
#define STAT_UP               3

/* ********************** types ********************* */

typedef struct {
  SEQUENCE_Message sequence;
  char variable[0];
} SequencedBuffer;

/**
 * Type of the connection table.
 **/
typedef struct BufferEntry_ {
  /** mutex guarding access to this BE */
  Mutex sem;
  /** Session for the connection */
  Session session;
  /** How much do we trust the host? signed
      because that makes many operations that go 
      negative easier. Of course, negative trust makes
      no sense. */
  int trust;
  /** the current session key */
  SESSIONKEY skey;
  /** at which time was the sessionkey created (by whichever party) */
  time_t created;
  /** is this host alive? timestamp of the time of the last-active point */
  cron_t isAlive; 
  /** what is the timeout of a cron-job that will flush this buffer? */
  cron_t flushTime;
  /**  Status of the connection (STAT_XXX) **/
  unsigned int status;
  /** last sequence number received (!) on this connection (highest) */
  unsigned int lastSequenceNumberReceived;
  /** bit map indicating which of the 32 sequence numbers before the last
      were received (good for accepting out-of-order packets and
      estimating reliability of the connection) */
  unsigned int lastPacketsBitmap;
  /** priority for flushing/sending this buffer, max(priorities of packets in buffer) */
  unsigned int prio; /* what is the max. priority in this buffer */
 /** number of bytes used in the buffer */
  unsigned short len;
  /** contents of the buffer */
  SequencedBuffer * buffer;
  /** a hash collision overflow chain **/
  struct BufferEntry_ * overflowChain;

} BufferEntry;

/* ********************* Methods called from "node" ********************** */

/**
 * Initialize this module.
 **/
void initConnection();

/**
 * Shutdown the connection module.
 **/
void doneConnection();

/**
 * Call this method periodically to scan data/hosts for new hosts.
 **/
void cronScanDirectoryDataHosts(void * unused);

/**
 * For debugging.
 **/
void printConnectionBuffer();

/* ********************** Send-interface ****************************** */

/**
 * Append a message to the current buffer.
 **/
void appendToBuffer(BufferEntry * be,
		    p2p_HEADER * message,
		    unsigned int maxdelay);

/**
 * Compute the hashtable index of a host id.
 **/
unsigned int computeIndex(const HostIdentity * hostId);

/**
 * Look for a host in the table. Also aquires the lock!
 * @returns the entry of the host in the table, NULL if not present
 **/
BufferEntry * lookForHost(const HostIdentity * hostId);


/**
 * Consider switching the transport mechanism used for contacting
 * the given node. This function is called when the handler handles
 * an encrypted connection. For example, if we are sending SMTP
 * messages to a node behind a NAT box, but that node has established
 * a TCP connection to us, it might just be better to send replies
 * on that TCP connection instead of keeping SMTP going.
 *
 * @param tsession the transport session that is for grabs
 * @param sender the identity of the other node
 **/
void considerTakeover(TSession * tsession,
		      HostIdentity * sender);

/**
 * Adding a host to the buffer.
 * @param force if YES, drop another host from the table if the slot is
 *              already in use. If NO, return NULL if the slot is busy.
 * @return the table entry for the host (no keyexchange performed so far)
 **/
BufferEntry * addHost(const HostIdentity * hostId,
		      int force);

/**
 * Register a callback method that should be invoked whenever a message
 * is about to be send that has more than minimumPadding bytes left
 * before maxing out the MTU. 
 * The callback method can then be used to add additional content
 * to the message (instead of the random noise that is added by
 * otherwise). Note that if the MTU is 0 (for streams), the
 * callback method will always be called with padding set to the
 * maximum number of bytes left in the buffer allocated for the
 * send.
 * @param minimumPadding how large must the padding be in order
 *   to call this method?
 * @param callback the method to invoke. The receiver is the
 *   receiver of the message, position is the reference to the
 *   first unused position in the buffer where GNUnet is building
 *   the message, padding is the number of bytes left in that buffer.
 *   The callback method must return the number of bytes written to
 *   that buffer (must be a positive number).
 * @return OK if the handler was registered, SYSERR on error
 **/
int registerSendCallback(const int minimumPadding,
			 BufferFillCallback callback);

/**
 * Unregister a handler that was registered with registerSendCallback.
 * @return OK if the handler was removed, SYSERR on error
 **/
int unregisterSendCallback(const int minimumPadding,
			   BufferFillCallback callback);

/**
 * Send a message to all directly connected nodes.
 *
 * @param message the message to send
 * @param priority how important is the message? The higher, the more important
 * @param maxdelay how long can we wait (max), in CRON-time (ms)
 **/
void broadcast(p2p_HEADER * message,
	       unsigned int priority,
	       unsigned int maxdelay);

/**
 * Send a message to a specific host (reply, enqueue)
 * @param message the message to send (unencrypted!), first 2 bytes give size
 * @param hostId the identity of the receiver
 * @param priority how important is the message?
 * @param maxdelay how long can we wait (max), in CRON-time (ms)
 **/
void unicast(const HostIdentity * hostId,
	     p2p_HEADER * message,
	     unsigned int priority,
	     unsigned int maxdelay);

/**
 * Type of a callback method on every buffer.
 * @param be the buffer entry
 * @param data context for callee
 **/
typedef void (*BufferEntryCallback)(BufferEntry * be,
				    void * data);

/**
 * Perform an operation for all connected hosts.
 * The BufferEntry structure is passed to the method.
 * No synchronization or other checks are performed.
 *
 * @param method the method to invoke (NULL for couting only)
 * @param data the second argument to the method
 * @return the number of connected hosts
 **/ 
int forAllConnectedHosts(BufferEntryCallback method,
			 void * data);


/**
 * Send a buffer; assumes that access to be is already 
 * synchronized.
 **/
void sendBuffer(BufferEntry * be);

/* ************************* encryption service ********************** */

/**
 * Decipher data comming in from a foreign host.
 * @param data the data to decrypt
 * @param size how big is data?
 * @param hostId the sender host that encrypted the data 
 * @param result where to store the decrypted data
 * @returns the size of the decrypted data, SYSERR on error
 **/
int decryptFromHost(void * data,
		    const unsigned short size,
		    const HostIdentity * hostId,
		    void * result);  

/**
 * We received a sign of life from this host.
 **/
void notifyPING(const HostIdentity * hostId);

/**
 * We received a sign of life from this host.
 **/
void notifyPONG(const HostIdentity * hostId);


#endif
/* end of connection.h */

