/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/
/**
 * @file util/configuration.c
 * @author Christian Grothoff
 *
 * Configuration management. This file provides code to parse a
 * configuration file (uses OpenSSL) and access the data stored in
 * it. It also provides methods to override options which can be used
 * to add command-line options to the configuration API.
 **/

#include "config.h"
#include <openssl/conf.h>
#include <openssl/err.h>

#include <string.h>

#include "util/configuration.h"
#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/storage.h"
#include "util/semaphore.h"

/**
 * Type for user-defined configuration entries.
 **/
typedef struct UserConfStruct {
  char * section;
  char * option;
  char * stringValue;
  int intValue;
  struct UserConfStruct * next;
} UserConf;

/**
 * GNUnet configuration (OpenSSL datastructure)
 **/
static LHASH * sslConfig = NULL;

/**
 * The filename of the config (for re-reading!)
 **/
static char * configuration_filename = NULL;

/**
 * Lock to access configuration (we may receive an
 * update signal at any time!)
 **/
static Mutex configLock;

/**
 * List of run-time options (not in the configuration
 * file, but for example from the command line).
 **/
static UserConf * userconfig = NULL;

/**
 * The command line strings (rest)
 **/ 
static char ** values;
static int valuesCount;

/**
 * Obtain a filename from the given section and option.  If the
 * filename is not specified, die with the given error message (do not
 * die if errMsg == NULL). 
 *
 * @param section from which section, may not be NULL
 * @param option which option, may not be NULL
 *
 * @param errMsg the errormessage, should contain two %s tokens for
 * the section and the option.
 *
 * @return the specified filename (caller must free), or NULL if no
 * filename was specified and errMsg == NULL
 **/
char * getFileName(char * section,
		   char * option,
		   const char * errMsg) {
  char * fn;  
  char * fnExpand;

  fn = getConfigurationString(section, 
			      option);
  if (fn == NULL) {
    if (errMsg == NULL)
      return NULL;
    else
      errexit(errMsg,
	      section,
	      option);
  }
  fnExpand = expandFileName(fn);   
  xfree(fn,
	"freeing string from getConfigurationString");
  return fnExpand;
}

/**
 * Read the specified configuration file. The previous
 * configuration will be discarded if this method is 
 * invoked twice.
 * @param fn the name of the configuration file (will
 *  be expanded)
 **/
void readConfiguration() {
  long errorLine; /* in which line did an error occur */

  /* getFileName aquires the mutex, so we better do this first */
  char * cfgName 
    = getFileName("FILES",
		  "gnunet.conf",
		  "No configuration file specified? Did you call readConfiguration before initConfiguration?");
  
  MUTEX_LOCK(&configLock);
  if (configuration_filename != NULL)
    xfree(configuration_filename,
	  "old config filename");
  if (sslConfig != NULL)
    CONF_free(sslConfig);
  configuration_filename 
    = cfgName;
  if (!assertIsFile(configuration_filename))
    errexit("Configuration file %s is not a file (or does not exist)\n", 
	    configuration_filename);
  sslConfig = CONF_load(NULL, 
			configuration_filename, 
			&errorLine);
  if (sslConfig == NULL) {
    ERR_load_crypto_strings();
    LOG(LOG_FAILURE,
	"FAILURE: Unable to load configuration from %s. Error at line %ld.\n",
	configuration_filename,
	errorLine);
    ERR_print_errors_fp(stderr);
    errexit("Bad configuration file %s.\n", 
	    configuration_filename);
  }
  MUTEX_UNLOCK(&configLock);
}

/**
 * This method must be called first! It reads
 * the config file and makes everything else
 * possible.
 **/
void initConfiguration() {
  MUTEX_CREATE(&configLock);
  setConfigurationString("FILES",
			 "gnunet.conf",
			 "~/.gnunet/gnunet.conf");
}

/**
 * This method may be called at last to clean up.
 * Afterwards everything but initConfiguration will result
 * in errors...
 **/
void doneConfiguration() {
  LOG(LOG_DEBUG,
      "DEBUG: in doneConfiguration\n");
  if (sslConfig != NULL)
    CONF_free(sslConfig); /* << segfaults here!? */
  sslConfig = NULL;
  if (configuration_filename != NULL)
    xfree(configuration_filename,
	  "doneConfiguration: config filename");
  MUTEX_DESTROY(&configLock);
}


/**
 * Obtain a string from the configuration.
 * @param section from which section, may not be NULL
 * @param option which option, may not be NULL
 * @return a freshly allocated string, caller must free!
 *   Note that the result can be NULL if the option is not set.
 **/
char * getConfigurationString(char * section,
			      char * option) {
  UserConf * pos;
  char * retval;

  if ( (section == NULL) || (option == NULL) )
    errexit("getConfigurationString called with section or option being NULL!\n");
  MUTEX_LOCK(&configLock);
  pos = userconfig;
  while (pos != NULL) {
    if ( (strcmp(section, pos->section) == 0) &&
	 (strcmp(option, pos->option) == 0) ) {
      if (pos->stringValue != NULL)
	retval = strdup(pos->stringValue);
      else
	retval = NULL; 
      MUTEX_UNLOCK(&configLock);
      return retval;
    }
    pos = pos->next;
  }
  retval = NULL;
  if (sslConfig != NULL)
    retval = CONF_get_string(sslConfig,
			     section,
			     option);
  if (retval != NULL)
    retval = strdup(retval);
  MUTEX_UNLOCK(&configLock);
  return retval;
}

/**
 * Check if a string in the configuration matches a given value.  This
 * method should be preferred over getConfigurationString since this
 * method can avoid making a copy of the configuration string that
 * then must be freed by the caller.
 *
 * @param section from which section, may not be NULL
 * @param option which option, may not be NULL
 * @param value the value to compare against
 * @return YES or NO
 **/
int testConfigurationString(char * section,
			    char * option,
			    const char * value) {
  UserConf * pos;
  char * c;

  if ( (section == NULL) || (option == NULL) )
    errexit("testConfigurationString called with section or option being NULL!\n");
  MUTEX_LOCK(&configLock);
  pos = userconfig;
  while (pos != NULL) {
    if ( (strcmp(section, pos->section) == 0) &&
	 (strcmp(option, pos->option) == 0) ) {
      if ( (pos->stringValue == NULL) &&
	   (value == NULL) ) {
	MUTEX_UNLOCK(&configLock);
	return YES;
      }
      if ( ( (pos->stringValue != NULL) ||
	     (value != NULL) ) && 
	   ( (pos->stringValue == NULL) ||
	     (value == NULL) ) ) {
	MUTEX_UNLOCK(&configLock);
	return NO;
      }
      if ( 0 == strcmp(pos->stringValue, value) ) {
	MUTEX_UNLOCK(&configLock);
	return YES;
      } else {
	MUTEX_UNLOCK(&configLock);
	return NO;
      }
    }
    pos = pos->next;
  }
  c = NULL;
  if (sslConfig != NULL)
    c = CONF_get_string(sslConfig,
			section,
			option);
  if ( (c == NULL) &&
       (value == NULL) ) {
    MUTEX_UNLOCK(&configLock);
    return YES;
  }
  if ( (c == NULL) ||
       (value == NULL) ) {
    MUTEX_UNLOCK(&configLock);
    return NO;
  }
  if ( 0 == strcmp(c, value) ) {
    MUTEX_UNLOCK(&configLock);
    return YES;
  } else {
    MUTEX_UNLOCK(&configLock);
    return NO;
  }
}

/**
 * Obtain an int from the configuration.
 *
 * @param section from which section, may not be NULL
 * @param option which option, may not be NULL
 * @return 0 if no option is specified
 **/
int getConfigurationInt(char * section,
			char * option) {
  UserConf * pos;
  int retval;

  if ( (section == NULL) || (option == NULL) )
    errexit("getConfigurationInt called with section or option being NULL!\n");
  MUTEX_LOCK(&configLock);
  pos = userconfig;
  while (pos != NULL) {
    if ( (strcmp(section, pos->section) == 0) &&
	 (strcmp(option, pos->option) == 0) ) {
      retval = pos->intValue;
      MUTEX_UNLOCK(&configLock);
      return retval;
    }
    pos = pos->next;
  }
  retval = 0;
  if (sslConfig != NULL)
    retval = CONF_get_number(sslConfig,
			     section,
			     option);
  MUTEX_UNLOCK(&configLock);
  return retval;
}

/**
 * Set an option.
 * @param section from which section, may not be NULL
 * @param option which option, may not be NULL
 * @param value the value to use, may be NULL
 * @return the previous value (or NULL if none),
 *     caller must free!
 **/
char * setConfigurationString(char * section,
			      char * option,
			      const char * value) {
  UserConf * pos;
  UserConf * prev; 
  char * res;

  if ( (section == NULL) || (option == NULL) )
    errexit("setConfigurationString called with section or option being NULL!\n");
  MUTEX_LOCK(&configLock);
  prev = NULL;
  pos = userconfig;
  while (pos != NULL) {
    if ( (strcmp(section, pos->section) == 0) &&
	 (strcmp(option, pos->option) == 0) ) {
      res = pos->stringValue;
      if (value != NULL)
	pos->stringValue = strdup(value);
      else
	pos->stringValue = NULL;
      MUTEX_UNLOCK(&configLock);
      return res;
    }
    prev = pos;
    pos = pos->next;
  }
  if (prev == NULL) {
    userconfig = xmalloc(sizeof(UserConf),
			 "new config entry");
    pos = userconfig;
  } else {
    prev->next = xmalloc(sizeof(UserConf),
			 "new config entry");
    pos = prev->next;
  }
  pos->section = strdup(section);
  pos->option = strdup(option);
  if (value != NULL)
    pos->stringValue = strdup(value);
  else 
    pos->stringValue = NULL;
  pos->intValue = 0;
  pos->next = NULL;
  res = NULL;
  if (sslConfig != NULL)
    res = CONF_get_string(sslConfig,
			  section,
			  option);
  if (res != NULL)
    res = strdup(res);
  MUTEX_UNLOCK(&configLock);
  return res;    
}

/**
 * Set an option.
 * @param section from which section, may not be NULL
 * @param option which option, may not be NULL
 * @param value the value to use
 * @return the previous value (or 0 if none)
 **/
int setConfigurationInt(char * section,
			char * option,
			const int value) {
  UserConf * pos;
  UserConf * prev; 
  int res;

  if ( (section == NULL) || (option == NULL) )
    errexit("setConfigurationInt called with section or option being NULL!\n");
  MUTEX_LOCK(&configLock);
  prev = NULL;
  pos = userconfig;
  while (pos != NULL) {
    if ( (strcmp(section, pos->section) == 0) &&
	 (strcmp(option, pos->option) == 0) ) {
      res = pos->intValue;
      pos->intValue = value;
      MUTEX_UNLOCK(&configLock);
      return res;
    }
    prev = pos;
    pos = pos->next;
  }
  if (prev == NULL) {
    userconfig = xmalloc(sizeof(UserConf),
			 "new config entry");
    pos = userconfig;
  } else {
    prev->next = xmalloc(sizeof(UserConf),
			 "new config entry");
    pos = prev->next;
  }
  pos->section = strdup(section);
  pos->option = strdup(option);
  pos->stringValue = NULL;
  pos->intValue = value;
  pos->next = NULL;
  res = 0;
  if (sslConfig != NULL)
    res = CONF_get_number(sslConfig,
			  section,
			  option);
  MUTEX_UNLOCK(&configLock);
  return res; 
}

/**
 * Get the command line strings (the ones
 * remaining after getopt-style parsing).
 * @param value the values
 + @return the number of values
 **/
int getConfigurationStringList(char *** value) {
  *value = values;
  return valuesCount;
}

/**
 * Set the list of command line options (remainder
 * after getopt style parsing).
 * @param value the values 
 + @param count the number of values
 **/
void setConfigurationStringList(char ** value,
				int count) {
  values = value;
  valuesCount = count;
}



/* end of configuration.c */
