/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file util/logging.c
 * @author Christian Grothoff
 *
 * This file contains basic logging mechanisms, with log-levels,
 * logging to file or stderr and with or without time-prefixing.
 **/

#include "util/logging.h"
#include "util/storage.h"
#include "util/xmalloc.h"
#include "util/semaphore.h"
#include "util/returnvalues.h"
#include "util/configuration.h"

static FILE * logfile = NULL;
static int loglevel__ = LOG_WARNING;
static int timestamping = NO;
static Mutex logMutex;

/**
 * Initialize the logging module.
 **/
void initLogging() {
  MUTEX_CREATE(&logMutex);
  resetLogging();
}

/**
 * Shutdown the logging module.
 **/
void doneLogging() {
  if ( (logfile != NULL) &&
       (logfile != stderr) )
    fclose(logfile);
  logfile = NULL;
  loglevel__ = 0;
  MUTEX_DESTROY(&logMutex);
}

/**
 * Convert a textual description of a loglevel into an int.
 **/
int getLogLevel() {
  return loglevel__;
}

void breakpoint() {
  LOG(LOG_FAILURE,
      "__BREAK__\n");
}
 
/**
 * Convert a textual description of a loglevel into an int.
 **/
static int getLoglevel(char * log) {
  if (log == NULL)
    errexit("LOGLEVEL specified is NULL, that's not ok.\n");
  if (0 == strcmp(log, "NOTHING"))
    return LOG_NOTHING;
  if (0 == strcmp(log, "FATAL"))
    return LOG_FATAL;
  if (0 == strcmp(log, "ERROR"))
    return LOG_ERROR;
  if (0 == strcmp(log, "FAILURE"))
    return LOG_FAILURE;
  if (0 == strcmp(log, "WARNING"))
    return LOG_WARNING;
  if (0 == strcmp(log, "MESSAGE"))
    return LOG_MESSAGE;
  if (0 == strcmp(log, "INFO"))
    return LOG_INFO;
  if (0 == strcmp(log, "DEBUG"))
    return LOG_DEBUG;
  if (0 == strcmp(log, "CRON"))
    return LOG_CRON;
  if (0 == strcmp(log, "EVERYTHING"))
    return LOG_EVERYTHING;
  errexit("invalid loglevel specified: %s (did you use all-caps?)\n",
	  log);
  return 42; /* can not happen */
}

/**
 * Re-read the loggig configuration.
 * Call on SIGHUP if the configuration file has changed.
 **/
void resetLogging() {
  char * loglevelname;
  char * logfilename;

  MUTEX_LOCK(&logMutex);
  loglevelname
    = getConfigurationString("GNUNETD",
			     "LOGLEVEL");
  loglevel__ 
    = getLoglevel(loglevelname); /* will errexit if loglevel == NULL */
  xfree(loglevelname,
	"freeing string of loglevel");

  timestamping
    = testConfigurationString("GNUNETD",
			      "TIMESTAMPS",
			      "YES"); 
  logfilename
    = getConfigurationString("GNUNETD",
			     "LOGFILE");
  if (logfilename != NULL) {
    logfilename = expandFileName(logfilename);
    logfile = fopen(logfilename, "a+");
    xfree(logfilename, 
	  "resetLogging - logfilename");
  } else
    logfile = stderr;
  MUTEX_UNLOCK(&logMutex);
}

/**
 * Print the current time to logfile without linefeed
 **/
static void printTime() {
  if ( (timestamping == YES) &&
       (logfile !=NULL) ) {
    char timebuf[64];
    time_t timetmp;
    struct tm * tmptr;
 
    time(&timetmp);
    tmptr = localtime(&timetmp);
    strftime(timebuf, 
	     64, 
	     "%b %e %H:%M:%S ", 
	     tmptr);
    fputs(timebuf, 
	  logfile);
  }
}

/**
 * Log a debug message
 * @param format the string describing the error message
 **/
void LOG(int minLogLevel,
	 const char *format, ...) {
  va_list	args;  

  if (loglevel__ < minLogLevel)
    return;
  MUTEX_LOCK(&logMutex);
  va_start(args, format);
  if (logfile != NULL) {
    printTime();
    vfprintf(logfile, format, args);
    fflush(logfile);
  } else
    vfprintf(stderr, format, args);
  va_end(args);
  MUTEX_UNLOCK(&logMutex);
}

/**
 * errexit - log an error message and exit.
 *
 * @param format the string describing the error message
 **/
void errexit(const char *format, ...) {
  va_list args;

  /* NO locking here, we're supposed to die,
     and we don't want to take chances on that... */
  va_start(args, format);
  if (logfile != NULL) {
    printTime();
    vfprintf(logfile, format, args);
    fflush(logfile);
  } else
    vfprintf(stderr, format, args);
  va_end(args);
  breakpoint();
  exit(1);
}

/* end of logging.c */
