/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file util/semaphore.c
 * This file contains functions related to threading.
 * In particular, functions for mutexes, semaphores 
 * and thread creation are provided.
 **/

#include "config.h"
#include "util/semaphore.h"
#include "util/logging.h"
#include <errno.h>
#include <string.h>
#ifdef SOMEBSD
#include <pthread_np.h>
#endif

/**
 * This prototype is somehow missing in various Linux pthread
 * include files. But we need it and it seems to be available
 * on all pthread-systems so far. Odd.
 **/
extern int pthread_mutexattr_setkind_np(pthread_mutexattr_t *attr, int kind);

/* -------------------------------------------------------------------- */
/* The following functions replace standard library functions in that   */
/* they exit on any error returned from the system calls.  Saves us     */
/* from having to check each and every call above.                      */
/* -------------------------------------------------------------------- */

static int tw_pthread_cond_wait(pthread_cond_t * c, 
				pthread_mutex_t * m) {
  int return_value;
  
  if ((return_value = pthread_cond_wait (c, m)) != 0)
    errexit("FATAL: pthread_cond_wait returned %d (%s)\n",
	    return_value, strerror(return_value));  	    
  return 0;
}

static int tw_pthread_cond_signal (pthread_cond_t * c) {
  int return_value;

  if ((return_value = pthread_cond_signal (c)) != 0)
    errexit("FATAL: pthread_cond_signal returned %d (%s)\n",
	    return_value, strerror(return_value));  
  return 0;
}

/* ********************* public methods ******************* */

void create_mutex_(Mutex * mutex) {
  pthread_mutexattr_t attr;

  pthread_mutexattr_init(&attr);
#if USE_CHECKING_MUTEX
#ifdef LINUX
  pthread_mutexattr_setkind_np(&attr,
			       PTHREAD_MUTEX_ERRORCHECK_NP);  
#endif
#else
  pthread_mutexattr_setkind_np(&attr,
			       (int)PTHREAD_MUTEX_NORMAL);
#endif
  if ( pthread_mutex_init(&mutex->pthreadMutex, &attr) != 0)
    errexit("FATAL: creation of mutex failed (%s)",
	    strerror(errno));
}

void create_recursive_mutex_(Mutex * mutex) {
  pthread_mutexattr_t attr;

  pthread_mutexattr_init(&attr);
#ifdef LINUX
  pthread_mutexattr_setkind_np(&attr,  
			       PTHREAD_MUTEX_RECURSIVE_NP);
#elif SOMEBSD
  pthread_mutexattr_setkind_np(&attr,
			       PTHREAD_MUTEX_RECURSIVE);
#elif SOLARIS
  pthread_mutexattr_settype(&attr,
			    PTHREAD_MUTEX_RECURSIVE);
#endif
  if ( pthread_mutex_init(&mutex->pthreadMutex, &attr) != 0)
    errexit("FATAL: creation of recursive mutex failed");
}

void destroy_mutex_(Mutex * mutex) {
  pthread_mutex_destroy(&mutex->pthreadMutex);
}

void mutex_lock_(Mutex * mutex) {
  int return_value;
  
  if ((return_value = pthread_mutex_lock (&mutex->pthreadMutex)) != 0)
    errexit("FATAL: pthread_mutex_lock returned %d (%s)\n",
	    return_value, strerror(return_value));  
}

void mutex_unlock_(Mutex * mutex) {
  int return_value;

  if ((return_value = pthread_mutex_unlock(&mutex->pthreadMutex)) != 0)
    errexit("FATAL: pthread_mutex_unlock returned %d (%s)\n",
	    return_value, strerror(return_value));  
}

/**
 * function must be called prior to semaphore use -- handles
 * setup and initialization.  semaphore destroy (below) should
 * be called when the semaphore is no longer needed.
 **/
Semaphore * new_semaphore (int value) {
  Semaphore * s = xmalloc(sizeof(Semaphore),
			  "new_semaphore: semaphore");
  s->v = value;
  MUTEX_CREATE(&(s->mutex));

  if (pthread_cond_init (&(s->cond), NULL) != 0)
    errexit("FATAL: pthread_cont_init failed\n");
  return s;
}

void free_semaphore(Semaphore * s) {
  MUTEX_DESTROY(&(s->mutex));
  pthread_cond_destroy(&(s->cond));
  xfree(s, "free_semaphore");
}

/**
 * function increments the semaphore and signals any threads that
 * are blocked waiting a change in the semaphore.
 **/
int semaphore_up (Semaphore * s) {
  int value_after_op;

  MUTEX_LOCK(&(s->mutex));  
  (s->v)++;
  value_after_op = s->v;  
  MUTEX_UNLOCK(&(s->mutex));
  tw_pthread_cond_signal (&(s->cond));  
  return (value_after_op);
}

/**
 * function decrements the semaphore and blocks if the semaphore is
 * <= 0 until another thread signals a change.
 **/
int semaphore_down (Semaphore * s) {
  int value_after_op;
  
  MUTEX_LOCK(&(s->mutex));
  while (s->v <= 0) {
    tw_pthread_cond_wait (&(s->cond), &(s->mutex.pthreadMutex));
  }
  (s->v)--;
  value_after_op = s->v;
  MUTEX_UNLOCK(&(s->mutex));  
  return (value_after_op);
}

/**
 * Create a thread. Use this method instead of pthread_create since
 * BSD may only give a 1k stack otherwise.
 *
 * @param stackSize the size of the stack of the thread in bytes. 
 *        Note that if the stack overflows, some OSes (seen under BSD) 
 *        will just segfault and gdb will give a messed-up stacktrace.
 * @return see pthread_create 
 **/
int PTHREAD_CREATE(pthread_t * handle,
		   void * (*main)(void *),
		   void * arg,
		   size_t stackSize) {
  pthread_attr_t stack_size_custom_attr;

  pthread_attr_init(&stack_size_custom_attr);
  pthread_attr_setstacksize(&stack_size_custom_attr,
			    stackSize);
  return pthread_create(handle, 
			&stack_size_custom_attr,
			main, 
			arg);			    
}



/* end of semaphore.c */
